/*
 * detect.c - libjcode $B%3!<%I<+F0H=Dj%k!<%A%s(B
 *                (C) Kuramitsu Kimio, Tokyo Univ. 1996-97 
 *
 * Ken Lunde $BCx(B $B!VF|K\8l>pJs=hM}!W(B $B$H(B jconv.c $B$r;29M$K:n$j(B 
 * $B$^$7$?!#(B
 * CGI $B%W%m%0%i%`$G$NMxMQ$r9M$($F!"$R$i$,$J!"$+$?$+$J$NJB$S(B
 * $B$K$h$k(BSJIS, EUC $B$N?dDj%"%k%4%j%:%`Ey$r2C$($F$"$j$^$9!#(B
 *
 */

#include <stdio.h>
#include "config.h"

/* ------------------------------------------------------ $B4A;z%3!<%I$N<1JL(B --*/

int _detect(unsigned char *str, int expected)
{
  register int c;

  while((c = (int)*str)!= '\0') {

    /* JIS $B%3!<%I$NH=Dj(B */
    if(c == ESC) {
      if((c = (int)*(++str)) == '\0') return expected; 
      if (c == '$') {
	if((c = (int)*(++str)) == '\0') return expected; 
	/* ESC $ B --> $B?7(BJIS 
	   ESC $ @ --> $B5l(BJIS */
	if (c == 'B' || c == '@') return JIS;
      }
#ifdef NECKANJI
      if (c == 'K')
	return NEC;    /* ESC K --> NEC JIS */
#endif
      str++;
      continue;
    }
    
    /* SJIS $B$K0l0U$K7hDj(B */
    if ((c >= 129 && c <= 141) || (c >= 143 && c <= 159))
        return SJIS;
    /* SS2 */
    if (c == SS2) {
      if((c = (int)*(++str)) == '\0') return expected; 
      if ((c >= 64 && c <= 126) || (c >= 128 && c <= 160) || 
	  (c >= 224 && c <= 252)) return SJIS;
      if (c >= 161 && c <= 223) expected = EUCORSJIS;
      str++;
      continue;
    }

    if (c >= 161 && c <= 223) {
      if((c = (int)*(++str)) == '\0') return expected; 
      
      if (c >= 240 && c <= 254)
	return EUC;
      if (c >= 161 && c <= 223) {
	expected = EUCORSJIS;
	str++;
	continue;
      }
      if (c <= 159) return SJIS;
      if (c >= 240 && c <= 254) return EUC;
     
      if (c >= 224 && c <= 239) {
	expected = EUCORSJIS;
	while (c >= 64) {
	  if (c >= 129) {
	    if (c <= 141 || (c >= 143 && c <= 159))
	      return SJIS;
	    else if (c >= 253 && c <= 254) {
	      return EUC;
	    }
	  }
	  if((c = (int)*(++str)) == '\0') return EUCORSJIS; 
	}
	str++;
	continue;
      }
      
      if (c >= 224 && c <= 239) {
        if((c = (int)*(++str)) == '\0') return expected; 
        if ((c >= 64 && c <= 126) || (c >= 128 && c <= 160))
          return SJIS;
        if (c >= 253 && c <= 254) return EUC;
        if (c >= 161 && c <= 252)
          expected = EUCORSJIS;
      }
    }
    str++;
  }
  return expected;
}

/* ---------------------------------------------------SJIS $B$+(B EUC $B$NH=Dj(B -- */
/* $BF|K\8l$NFCD'(B($B$R$i$,$J!"%+%?%+%J(B)$B$+$iH=Dj$9$k$?$a!"(B
   $BH=Dj%_%9$,H/@8$9$k>l9g$b$"$j$^$9!#(B */

int _detect_euc_or_sjis(unsigned char *str) {
  int c1, c2 = -1;
  int expected = EUCORSJIS; /* $B%G%U%)%k%H$G(BSJIS */

  while((c1 = (int)*str) != '\0') {
    if(c1 == 0xa4 || c1 == 0xa5) {
      str++;
      if(*str == '\0') break; else str++;
      if((c2 = (int)*str) == '\0') break;
      if(c2 == 0xa4 || c2 == 0xa5) {
	expected = EUC;
	return EUC;
      }
    }
    str++;
  }
  return expected;
}

int _cache_euc_or_sjis(int code)
{
  static int cache = SJIS;     /* $B%G%U%)%k%H$O!"(BSJIS */
  if(code == EUC || code == SJIS ) cache = code;
  if(code == EUCORSJIS) return cache; else return code;
}

/* ---------------------------------------------------------- Public $B4X?t(B -- */


#ifdef DEBUG
void printDetectCode(int detected) {
  switch(detected) {
  case ASCII:
    printf("ASCII/JIS-Roman characters(94 printable)\n");
    break;
  case JIS:
    printf("JIS(iso-2022-jp)\n");
    break;
  case EUC:
    printf("EUC(x-euc-jp)\n");
    break;
  case SJIS:
    printf("SJIS(x-sjis)\n");
    break;
  case NEW:
    printf("JIS X 0208-1990\n");
    break;
  case OLD:
    printf("JIS X 0208-1978\n");
    break;
  case EUCORSJIS:
    printf("EUC or SJIS\n");
    break;
  default:
    printf("Another Codes!!\n");
  }
}
#endif

int detectKanjiCode(char *str)
{
  static int detected = ASCII;

  /* JIS, EUC, SJIS, EUCORSJIS $B$NH=Dj(B */
  detected = _detect((unsigned char *)str, ASCII);

  /* $B?7(BJIS, $B5l(BJIS, NEC JIS $B$N>l9g!"(BJIS $B$KJQ99$9$k(B */
  if(detected == NEW || detected == OLD || detected == NEC)
    return JIS;

#ifndef NO_NARABI_CHECK
  /* SJIS $B$+(B EUC $B$N6hJL$r%+%J$NJB$S$+$i?dDj$9$k(B */
  if(detected == EUCORSJIS)
    detected = _detect_euc_or_sjis((unsigned char *)str);
#endif

  /* SJIS $B$+(B EUC $B$NH=Dj$K5g$7$?$iA02s$NH=Dj$K=>$&(B */
  detected = _cache_euc_or_sjis(detected);

  return detected;
}
