/* 
 *  libjcode.c -- $B4A;zJQ49%i%$%V%i%j(B    1.0 $BHG(B
 *                (C) Kuramitsu Kimio, Tokyo Univ. 1996-97
 *
 *  $B$3$N%i%$%V%i%j$O!"(BCGI Programming with C and Perl $B$N$?$a$K(B
 *  Ken Lunde $BCx(B $B!VF|K\8l>pJs=hM}!W(B (O'llery) $B$r;29M$K$7$F!"(B
 *  $B%9%H%j!<%`MQ$@$C$?(Bjconv.c $B$r!"%9%H%j%s%0BP1~$K$7$F%i%$%V%i%j2=(B
 *  $B$7$^$7$?!#(B 
 *  $B$?$@$7!"(BCGI (INTERNET)$B$G$NMxMQ$r9M$($F!"JQ99$7$F$"$j$^$9!#(B
 */
 
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "config.h"
 
extern int detectKanjiCode(char *str);
 
#define CHAROUT(ch) *str2 = (unsigned char)(ch); str2++;

/* --------------------------------------- JIS(ISO-2022) $B%3!<%I$X@Z$jBX$((B -- */

unsigned char *_to_jis(unsigned char *str) {
  *str = (unsigned char)ESC; str++;
  *str = (unsigned char)'$'; str++;
  *str = (unsigned char)'B'; str++;
  return str;
}

/* ----------------------------------------------- ASCII $B%3!<%I$X@Z$jBX$((B -- */

/* ESC ( B $B$H(B ESC ( J $B$N0c$$!#(B
   $BK\Mh$O!"(B ESC ( J $B$,@5$7$$(BJIS-Roman $BBN7O$G$"$k$,!"(B
   $B%$%s%?!<%M%C%H$N>e$G$O!"1Q?t;z$O(BASCII $B$NJ}$,<+A3$+$H;W$o$l$k!#(B
   \ $B5-9f$H(B ~$B5-9f$,0c$&$@$1$G$"$k!#(B */

unsigned char *_to_ascii(unsigned char *str) {
  *str = (unsigned char)ESC; str++;
  *str = (unsigned char)'('; str++;
  *str = (unsigned char)'B'; str++;
  return str;
}

/* -------------------------------------- JIS $B%3!<%I(B $B$r(B SJIS$B$H$7$F%7%U%H(B -- */

void _jis_shift(int *p1, int *p2)
{
  unsigned char c1 = *p1;
  unsigned char c2 = *p2;
  int rowOffset = c1 < 95 ? 112 : 176;
  int cellOffset = c1 % 2 ? (c2 > 95 ? 32 : 31) : 126;

  *p1 = ((c1 + 1) >> 1) + rowOffset;
  *p2 += cellOffset;
}

/* --------------------------------- SJIS $B%3!<%I$r(BJIS $B%3!<%I$H$7$F%7%U%H(B -- */

void _sjis_shift(int *p1, int *p2)
{
  unsigned char c1 = *p1;
  unsigned char c2 = *p2;
  int adjust = c2 < 159;
  int rowOffset = c1 < 160 ? 112 : 176;
  int cellOffset = adjust ? (c2 > 127 ? 32 : 31) : 126;

  *p1 = ((c1 - rowOffset) << 1) - adjust;
  *p2 -= cellOffset;
}

/* ---------------------------------------------- SJIS $BH>3Q$rA43Q$KJQ49(B -- */
#define HANKATA(a)  (a >= 161 && a <= 223)
#define ISMARU(a)   (a >= 202 && a <= 206)
#define ISNIGORI(a) ((a >= 182 && a <= 196) || (a >= 202 && a <= 206) || (a == 179))

static int stable[][2] = {
    {129,66},{129,117},{129,118},{129,65},{129,69},{131,146},{131,64},
    {131,66},{131,68},{131,70},{131,72},{131,131},{131,133},{131,135},
    {131,98},{129,91},{131,65},{131,67},{131,69},{131,71},{131,73},
    {131,74},{131,76},{131,78},{131,80},{131,82},{131,84},{131,86},
    {131,88},{131,90},{131,92},{131,94},{131,96},{131,99},{131,101},
    {131,103},{131,105},{131,106},{131,107},{131,108},{131,109},
    {131,110},{131,113},{131,116},{131,119},{131,122},{131,125},
    {131,126},{131,128},{131,129},{131,130},{131,132},{131,134},
    {131,136},{131,137},{131,138},{131,139},{131,140},{131,141},
    {131,143},{131,147},{129,74},{129,75}};

unsigned char *_sjis_han2zen(unsigned char *str, int *p1, int *p2)
{
  register int c1, c2;

  c1 = (int)*str; str++;
  *p1 = stable[c1 - 161][0];
  *p2 = stable[c1 - 161][1];

  /* $BBy2;!"H>By2;$N=hM}(B */
  c2 = (int)*str;
  if (c2 == 222 && ISNIGORI(c1)) {
    if ((*p2 >= 74 && *p2 <= 103) || (*p2 >= 110 && *p2 <= 122))
      (*p2)++;
    else if (*p1 == 131 && *p2 == 69)
      *p2 = 148;
    str++;
  }

  if (c2 == 223 && ISMARU(c1) && (*p2 >= 110 && *p2 <= 122) ) {
    *p2 += 2;
    str++;
  }
  return str++;
}

/* -------------------------------------------------- SJIS $B%3!<%I$rJQ49(B -- */

#define SJIS1(A)    ((A >= 129 && A <= 159) || (A >= 224 && A <= 239))
#define SJIS2(A)    (A >= 64 && A <= 252)

void _shift2seven(unsigned char *str, unsigned char *str2)
{
  int p1,p2,esc_in = FALSE;

  while ((p1 = (int)*str) != '\0') {

    if (SJIS1(p1)) {
      if((p2 = (int)*(++str)) == '\0') break;
      if (SJIS2(p2)) {
        _sjis_shift(&p1,&p2);
        if (!esc_in) {
          esc_in = TRUE;
          str2 = _to_jis(str2);
        }
      }
      CHAROUT(p1);
      CHAROUT(p2);
      str++;
      continue;
    }

#ifdef NO_HANKAKU_SJIS
    /* $BH>3Q(B SJIS $B$O!"6/@)E*$KA43Q$KJQ$($k(B */
    if (HANKATA(p1)) {
      str = _sjis_han2zen(str, &p1, &p2);
      _sjis_shift(&p1,&p2);
      if (!esc_in) {
        esc_in = TRUE;
        str2 = _to_jis(str2);
      }
      CHAROUT(p1);
      CHAROUT(p2);
      continue;
    }
#endif

    if (esc_in) {
      /* LF / CR $B$N>l9g$O!"@5>o$K%(%9%1!<%W%"%&%H$5$l$k(B */
      esc_in = FALSE;
      str2 = _to_ascii(str2);
    }
    CHAROUT(p1);
    str++;
  }

  if (esc_in)
    str2 = _to_ascii(str2);
  *str2='\0';
}

/* --------------------------------------------- SJIS $B$r(B EUC $B$KJQ49$9$k(B -- */

void _shift2euc(unsigned char *str, unsigned char *str2)
{
  int p1,p2;
  
  while ((p1 = (int)*str) != '\0') {
    if (SJIS1(p1)) {
      if((p2 = (int)*(++str)) == '\0') break;
      if (SJIS2(p2)) {
        _sjis_shift(&p1,&p2);
        p1 += 128;
        p2 += 128;
      }
      CHAROUT(p1);
      CHAROUT(p2);
      str++;
      continue;
    }

#ifdef NO_HANKAKU_SJIS
    /* $BH>3Q(B SJIS $B$O!"6/@)E*$KA43Q$KJQ$($k(B */
    if (HANKATA(p1)) {
      str = _sjis_han2zen(str,&p1,&p2);
      _sjis_shift(&p1,&p2);
      p1 += 128;
      p2 += 128;
      CHAROUT(p1);
      CHAROUT(p2);
      continue;
    }
#endif
    CHAROUT(p1);
    str++;
  }
  *str2='\0';
}

/* ------------------------------------------------- $BH>3Q(B SJIS $B$r<h$j=|$/(B -- */

void _shift_self(unsigned char *str, unsigned char *str2)
{
  int p1,p2;
  
  while ((p1 = (int)*str) != '\0') {
#ifdef NO_HANKAKU_SJIS
    /* $BH>3Q(B SJIS $B$O!"6/@)E*$KA43Q$KJQ$($k(B */
    if (HANKATA(p1)) {
      str = _sjis_han2zen(str, &p1, &p2);
      CHAROUT(p1);
      CHAROUT(p2);
      continue;
    }
#endif
    CHAROUT(p1);
    str++;
  }
  *str2='\0';
}

/* ------------------------------------------------------EUC $B$+$i(B JIS $B$X(B -- */

#define ISEUC(A)    (A >= 161 && A <= 254)

void _euc2seven(unsigned char *str, unsigned char *str2)
{
  int p1, p2, esc_in = FALSE;

  while ((p1 = (int)*str) != '\0') {

    if (p1 == LF || p1 == CR) {
      if (esc_in) {
        esc_in = FALSE;
        str2 = _to_ascii(str2);
      }
      CHAROUT(p1);
      str++;
      continue;
    }

    if (ISEUC(p1)) {
      if((p2 = (int)*(++str)) == '\0') break;
      if (ISEUC(p2)) {

	if (!esc_in) {
	  esc_in = TRUE;
	  str2 =_to_jis(str2);
	}

	CHAROUT(p1-128);
	CHAROUT(p2-128);
	str++;
	continue;
      }
    }

    if (esc_in) {
      esc_in = FALSE;
      str2 = _to_ascii(str2);
    }
    CHAROUT(p1);
    str++;
  }
  *str2='\0';
}

/* ------------------------------------------------ EUC $B$+$i(B SJIS $B$KJQ49(B -- */
 
void _euc2shift(unsigned char *str, unsigned char *str2)
{
  int p1,p2;

  while ((p1 = (int)*str) != '\0') {
    if (ISEUC(p1)) {
      if((p2 = (int)*(++str)) == '\0') break;
      if (ISEUC(p2)) {
	p1 -= 128;
        p2 -= 128;
        _jis_shift(&p1,&p2);
      }
      CHAROUT(p1);
      CHAROUT(p2);
      str++;
      continue;
    }

    CHAROUT(p1);
    str++;
  }
  *str2='\0';
}

/* -------------------------------------- ESC $B%7!<%1%s%9$r%9%-%C%W$9$k(B ----- */

unsigned char *_skip_esc(unsigned char *str, int *esc_in) {
  int c;
  
  c = (int)*(++str);
  if ((c == '$') || (c == '(')) str++;
  if ((c == 'K') || (c == '$')) *esc_in = TRUE;
  else *esc_in = FALSE;

  if(*str != '\0') str++;
  return str;
}


/* ----------------------------------------------- JIS $B$r(B SJIS $B$KJQ49$9$k(B -- */

void _seven2shift(unsigned char *str, unsigned char *str2)
{
  int p1, p2, esc_in = FALSE;

  while ((p1 = (int)*str) != '\0') {

    /* ESC$B%7!<%1%s%9$r%9%-%C%W$9$k(B */
    if (p1 == ESC) {
      str = _skip_esc(str, &esc_in);
      continue;
    }

    if (p1 == LF || p1 == CR) {
      if (esc_in) esc_in = FALSE;
    }

    if(esc_in) { /* ISO-2022-JP $B%3!<%I(B */
      if((p2 = (int)*(++str)) == '\0') break;

      _jis_shift(&p1, &p2);

      CHAROUT(p1);
      CHAROUT(p2);
    }else{       /* ASCII $B%3!<%I(B */
      CHAROUT(p1);
    }
    str++;
  }
  *str2 = '\0';
}

/* ------------------------------------------------ JIS $B$r(B EUC $B$KJQ49$9$k(B -- */

void _seven2euc(unsigned char *str, unsigned char *str2)
{
  int p1, esc_in = FALSE;

  while ((p1 = (int)*str) != '\0') {

    /* ESC$B%7!<%1%s%9$r%9%-%C%W$9$k(B */
    if (p1 == ESC) {
      str = _skip_esc(str, &esc_in);
      continue;
    }

    if (p1 == LF || p1 == CR) {
      if (esc_in) esc_in = FALSE;
    }

    if(esc_in) { /* ISO-2022-JP $B%3!<%I(B */
      CHAROUT(p1 + 128); 
      
      if((p1 = (int)*(++str)) == '\0') break;
      CHAROUT(p1 + 128);
    }else{       /* ASCII $B%3!<%I(B */
      CHAROUT(p1);
    }
    str++;
  }
  *str2 = '\0';
}

/* ------------------------------------------------------------------------ */
/* --------------------------------------------------------- Public $B4X?t(B -- */

#define STATIC_BUFFER BUFSIZ

unsigned char *_set_buffer(char *str) {
#ifdef STATIC_BUFFER
  static unsigned char buf[STATIC_BUFFER];
#else
  static unsigned char *buf;

  if((buf = (unsigned char *)malloc((strlen(str) + 1) * 3)) == NULL) {
    fprintf(stderr, "Can't malloc buffer\n");
    exit(2);
  }
#endif
  return buf;
}

char *_replace_buffer(unsigned char *buf) {
  unsigned char *p;
  char *str;

  if((str = (char *)malloc(strlen(buf) + 1)) == NULL) {
    fprintf(stderr, "Can't malloc string buffer\n");
    exit(2);
  }

  p = (unsigned char *)str;
   
  while(*buf) {
    *p = *buf; buf++; p++;
  }
  *p = '\0';
#ifndef STATIC_BUFFER
  free(buf);
#endif
  return str;
}

char *toStringJIS(char *str) {
  int detected;
  unsigned char *buf;

  detected = detectKanjiCode(str);

  if(detected == ASCII || detected == JIS)
    return str;

  buf = _set_buffer(str);

  switch(detected) {
  case SJIS :
    _shift2seven((unsigned char *)str, buf);
    break;
  case EUC :
    _euc2seven((unsigned char *)str, buf);
    break;
  default:
    return str;
    break;
  }

  return (char *)_replace_buffer(buf);
}

char *toStringEUC(char *str) {
  int detected;
  unsigned char *buf;

  detected = detectKanjiCode(str);

  if(detected == ASCII) return str;

  buf = _set_buffer(str);

  switch(detected) {
  case SJIS :
    _shift2euc((unsigned char *)str, buf);
    break;
  case JIS :
  case NEW : case OLD : case NEC :
     _seven2euc((unsigned char *)str, buf);
    break;
  default:
    return str;
    break;
  }

  return (char *)_replace_buffer(buf);
}

char *toStringSJIS(char *str) {
  int detected;
  unsigned char *buf;

  detected = detectKanjiCode(str);
  if(detected == ASCII) return str;
  
  buf = _set_buffer(str);

  switch(detected) {
  case NEW : case OLD : case NEC :
  case JIS :
    _seven2shift((unsigned char *)str, buf);
    break;
  case EUC :
    _euc2shift((unsigned char *)str, buf);
    break;
  case SJIS :  
  default:
    _shift_self((unsigned char *)str, buf);
  }

  return _replace_buffer(buf);
}


char *test(char *str) {
  static char buffer[BUFSIZ];
  
  _shift2seven((unsigned char *)str, (unsigned char *)buffer);
  
  return (char *)buffer;
}









